# -*- coding: utf-8 -*-
"""
Created on Sun Feb  9 11:12:20 2020
@author: mendez
"""

import numpy as np
from numpy import linalg as LA
from scipy.signal import firwin 
from scipy.sparse.linalg import svds
import matplotlib.pyplot as plt
from matplotlib.patches import Circle


# Generalized Decomposition
def Modal_Analysis_2(D,Psi,Sort):  
    """
    This function implements algorithm 2 in the notes.          
    :param D: Dataset Matrix
    :param PSI: Temporal Basis
    :param Sort: 1 for sorting 0 for not
    """  

    # Step 1: Prepare Projection
    P_Psi_s=np.dot(np.transpose(np.conjugate(Psi)),Psi)
    P_Psi=np.dot(np.conjugate(Psi),LA.inv(P_Psi_s))
    # Step 2: Project the data (line 1)
    D_psi=np.dot(D,P_Psi)           
    # Step 3: Loop for Phi and Sigma
    # Initialize output first
    R=Psi.shape[1] 
    Phi=np.zeros((D.shape[0],R),dtype=complex)
    Sigma=np.zeros(R,)     
    # Loop for normalization (line 3)
    for r in range(0,R):
     Sigma[r]=LA.norm(D_psi[:,r])
     Phi[:,r]=D_psi[:,r]/Sigma[r]  
     print('Computing Mode '+str(r+1)+'/'+str(R)) 
    # We add also the possibility to do sorting
    if Sort:
     # find indices for sorting in decreasing order   
     Indices=np.flipud(np.argsort(Sigma)) 
     # Sort The decomposition
     Sigma=Sigma[Indices] 
     Phi=Phi[:,Indices] 
     Psi=Psi[:,Indices] 
    return Phi, Sigma, Psi



# Generalized Decomposition
def Modal_Analysis_Ortho(D,Psi,Sort):   
    """
    This function implements algorithm 2 in the notes,
    assuming the the temporal structures are orthonormal          
    :param D: Dataset Matrix
    :param PSI: Temporal Basis
    :param Sort: 1 for sorting 0 for not    
    """  

    # Step 1: Prepare Projection
    P_Psi=np.conjugate(Psi)
    # Step 2: Project the data (line 1)
    D_psi=np.dot(D,P_Psi)           
    # Step 3: Loop for Phi and Sigma
    # Initialize output first
    R=Psi.shape[1] 
    Phi=np.zeros((D.shape[0],R),dtype=complex)
    Sigma=np.zeros(R,)     
    # Loop for normalization (line 3)
    for r in range(0,R):
     Sigma[r]=LA.norm(D_psi[:,r])
     Phi[:,r]=D_psi[:,r]/Sigma[r]  
     print('Computing Mode '+str(r+1)+'/'+str(R))    
    # We add also the possibility to do sorting
    if Sort:
     # find indices for sorting in decreasing order   
     Indices=np.flipud(np.argsort(Sigma)) 
     # Sort The decomposition
     Sigma=Sigma[Indices] 
     Phi=Phi[:,Indices] 
     Psi=Psi[:,Indices] 
 
    return Phi, Sigma, Psi





# mPOD Decomposition
def mPOD_A3(D,F_V,Nf,f_s):  
    """
    This function implements algorithm 2 in the notes.          
    :param D: Dataset Matrix
    :param F_V: Frequency Splitting Vector
    :param Nf: vector with order (must be odd!) of the kernels
    :param f_s: Sampling Frequency
    
    """
    ## Step 1: Compute K and K hat 
    K = np.dot(D.transpose(), D); 
    
    ## Step 2-3-4-5: MRA
    F_Bank_r = F_V*2/f_s #(f_s/2 mapped to 1)
    M=len(F_Bank_r) # Number of scales
    
    # Initialize some of the matrices of interest
    Psi_M=np.array([]) # 
    Lambda_M=np.array([])
    n_t=D.shape[1]; n_s=D.shape[0]
    # Tensor storing all the correlation Matrices
    Ks=np.zeros((n_t,n_t,M+1))

    
    for m in range(0,M):
     if m<1: # construct largest (Low PASS)
      h_A=firwin(Nf[m], F_Bank_r[m], window='hamming')
      # Prepare Transfer function (zero-phase-shift)
      print('Filtering Largest Scale')
      # Filter Largest scale
      K_L=Filt_K(K,h_A,n_s); Ks[:,:,m]=K_L # First large scale
      print('Diagonalizing Largest Scale')
      R_K=LA.matrix_rank(K_L, tol=None, hermitian=True)
      Psi_P, Lambda_P, _ = svds(K_L,R_K)
      Psi_M=Psi_P # In the first scale we take it as is
      Lambda_M=Lambda_P
     
      # Construct first band pass
      if M>1: # then this is band-pass
       h1d_H= firwin(Nf[m],\
      [F_Bank_r[m],F_Bank_r[m+1]],pass_zero=False) 
      else: #if M=1, then the first band pass is high pass
       h1d_H= firwin(Nf[m],F_Bank_r[m],pass_zero=False) 

      print('Filtering H Scale '+str(m+1)+'/'+str(M))
      K_H=Filt_K(K,h1d_H,n_s); Ks[:,:,m+1]=K_H # First band pass
      print('Diagonalizing H Scale '+str(m+1)+'/'+str(M))
      R_K=LA.matrix_rank(K_H, tol=None, hermitian=True)
      Psi_P, Lambda_P, _ = svds(K_H,R_K) # Diagonalize scale
      Psi_M=np.concatenate((Psi_M, Psi_P), axis=1) # append to the previous
      Lambda_M=np.concatenate((Lambda_M, Lambda_P), axis=0)
      
     
     elif m>0 and m<M-1: # 
     # This is the 1d Kernel for Band pass
      h1d_H= firwin(Nf[m],[F_Bank_r[m],F_Bank_r[m+1]],pass_zero=False) # Band-pass
      print('Filtering H Scale '+str(m+1)+'/'+str(M))
      K_H=Filt_K(K,h1d_H,n_s); Ks[:,:,m+1]=K_H # Intermediate band-pass
      print('Diagonalizing H Scale '+str(m+1)+'/'+str(M))
      R_K=LA.matrix_rank(K_H, tol=None, hermitian=True)
      Psi_P, Lambda_P, _ = svds(K_H,R_K) # Diagonalize scale
      Psi_M=np.concatenate((Psi_M, Psi_P), axis=1) # append to the previous
      Lambda_M=np.concatenate((Lambda_M, Lambda_P), axis=0)
    else:
      # This is the 1d Kernel for High pass
     h1d_H = firwin(Nf[m],F_Bank_r[m],pass_zero=False)
     print('Filtering H Scale '+str(m+1)+'/ '+str(M))
     K_H=Filt_K(K,h1d_H,n_s); Ks[:,:,m+1]=K_H # Last (high pass) scale
     print('Diagonalizing H Scale '+str(m+1)+'/ '+str(M))
     R_K=LA.matrix_rank(K_H, tol=None, hermitian=True)
     Psi_P, Lambda_P, _ = svds(K_H,R_K) # Diagonalize scale
     Psi_M=np.concatenate((Psi_M, Psi_P), axis=1) # append to the previous
     Lambda_M=np.concatenate((Lambda_M, Lambda_P), axis=0)
    return Psi_M,Ks # Return temporal basis + Correlation Scales




## first construct a function for the filtering
    

def Filt_K(K,h,n_s):
   """
    This function Filters K, assuming periodic BC, and a FIR
    with impulse response h
    
   """
   n_t=K.shape[1]; # size of K
   K_HAT=np.fft.fft2(K); # Spectra of K
   # Zero-Phase (anticausal) H_underline
   H_hat=np.fft.fft(h,n_t); H_hat_zs=H_hat*np.conj(H_hat);
   H_hat_r=np.array([H_hat_zs]) # Store as row vector
   H_cal=np.dot(np.ones((n_s,1)),H_hat_r)/np.sqrt(n_s)
   H_underline=np.dot(np.conj(np.transpose(H_cal)),H_cal)
   # Filter K
   K_filt=np.real(np.fft.ifft2(K_HAT*H_underline))
   return K_filt


## Plot Fields tutorial 1
   
def Plot_Field_TEXT_Jet(File):  
   """
   This function plots the vector field from the TR-PIV in Exercise 4.
      
    :param File: Name of the file to load          
   
   """
   # We first perform a zero padding
   Name=File
   # Read data from a file
   DATA = np.genfromtxt(Name) # Here we have the four colums
   Dat=DATA[1:,:] # Here we remove the first raw, containing the header
   nxny=Dat.shape[0] # is the to be doubled at the end we will have n_s=2 * n_x * n_y
   n_s=2*nxny
   ## 1. Reconstruct Mesh from file
   X_S=Dat[:,0];
   Y_S=Dat[:,1];
   # Reshape also the velocity components
   V_X=Dat[:,2] # U component
   V_Y=Dat[:,3] # V component
   # Put both components as fields in the grid
   Xg,Yg,Vxg,Vyg,Magn=Plot_Field_Jet(X_S,Y_S,V_X,V_Y,False,2,0.6)
   # Show this particular step
   fig, ax = plt.subplots(figsize=(5, 3)) # This creates the figure
   # Or you can plot it as streamlines
   plt.contourf(Xg,Yg,Magn)
   # One possibility is to use quiver
   STEPx=2
   STEPy=2
   plt.quiver(Xg[::STEPx,::STEPy],Yg[::STEPx,::STEPy],\
               Vxg[::STEPx,::STEPy],-Vyg[::STEPx,::STEPy],color='k') # Create a quiver (arrows) plot
    
   plt.rc('text', usetex=True)      # This is Miguel's customization
   plt.rc('font', family='serif')
   plt.rc('xtick',labelsize=16)
   plt.rc('ytick',labelsize=16)
   ax.set_aspect('equal') # Set equal aspect ratio
   ax.set_xlabel('$x[mm]$',fontsize=13)
   ax.set_ylabel('$y[mm]$',fontsize=13)
   ax.set_title('Tutorial 1: Impinging Jet',fontsize=16)
   ax.set_xticks(np.arange(0,40,10))
   ax.set_yticks(np.arange(10,30,5))
   ax.set_xlim([0,35])
   ax.set_ylim(10,29)
   ax.invert_yaxis() # Invert Axis for plotting purpose
   plt.tight_layout()
   plt.show()
   Name[len(Name)-12:len(Name)]+' Plotted'
   return n_s, Xg, Yg, Vxg, -Vyg, X_S, Y_S


def Plot_Field_Jet(X_S,Y_S,V_X,V_Y,PLOT,Step,Scale):
 # Number of n_X/n_Y from forward differences
   GRAD_X=np.diff(X_S) 
   #GRAD_Y=np.diff(Y_S);
   # Depending on the reshaping performed, one of the two will start with
   # non-zero gradient. The other will have zero gradient only on the change.
   IND_X=np.where(GRAD_X!=0); DAT=IND_X[0]; n_y=DAT[0]+1;
   nxny=X_S.shape[0] # is the to be doubled at the end we will have n_s=2 * n_x * n_y
   #n_s=2*nxny
   # Reshaping the grid from the data
   n_x=(nxny//(n_y)) # Carefull with integer and float!
   Xg=np.transpose(X_S.reshape((n_x,n_y)))
   Yg=np.transpose(Y_S.reshape((n_x,n_y))) # This is now the mesh! 60x114.
   Mod=np.sqrt(V_X**2+V_Y**2)
   Vxg=np.transpose(V_X.reshape((n_x,n_y)))
   Vyg=np.transpose(V_Y.reshape((n_x,n_y)))
   Magn=np.transpose(Mod.reshape((n_x,n_y)))
   if PLOT:
   # Show this particular step
#    fig, ax = plt.subplots(figsize=(8, 5)) # This creates the figure
    plt.contourf(Xg,Yg,Magn)
    # One possibility is to use quiver
    STEPx=Step
    STEPy=Step
    plt.quiver(Xg[::STEPx,::STEPy],Yg[::STEPx,::STEPy],\
               Vxg[::STEPx,::STEPy],-Vyg[::STEPx,::STEPy],color='k',scale=Scale) # Create a quiver (arrows) plot
    
    plt.rc('text', usetex=True)      # This is Miguel's customization
    plt.rc('font', family='serif')
    plt.rc('xtick',labelsize=16)
    plt.rc('ytick',labelsize=16)
#    ax.set_aspect('equal') # Set equal aspect ratio
#    ax.set_xlabel('$x[mm]$',fontsize=16)
#    ax.set_ylabel('$y[mm]$',fontsize=16)
#    ax.set_title('Velocity Field via TR-PIV',fontsize=18)
#    ax.set_xticks(np.arange(0,40,10))
#    ax.set_yticks(np.arange(10,30,5))
#    ax.set_xlim([0,35])
#    ax.set_ylim(10,29)
#    ax.invert_yaxis() # Invert Axis for plotting purpose
#    plt.show()
       
   return Xg,Yg,Vxg,Vyg,Magn  


## Plot Fields tutorial 1
   
def Plot_Field_TEXT_Cylinder(File,Name_Mesh):  
   """
   This function plots the vector field from the TR-PIV in Exercise 4.
      
    :param File: Name of the file to load          
   
   """
   # We first perform a zero padding
   Name=File
   # Read data from a file
   DATA = np.genfromtxt(Name) # Here we have the four colums
   Dat=DATA[1:,:] # Here we remove the first raw, containing the header
   nxny=Dat.shape[0] # is the to be doubled at the end we will have n_s=2 * n_x * n_y
   n_s=2*nxny
   ## 1. Reconstruct Mesh from file Name_Mesh
   DATA_mesh=np.genfromtxt(Name_Mesh);
   Dat_mesh=DATA_mesh[1:,:]
   X_S=Dat_mesh[:,0];
   Y_S=Dat_mesh[:,1];
   # Reshape also the velocity components
   V_X=Dat[:,0] # U component
   V_Y=Dat[:,1] # V component
   # Put both components as fields in the grid
   Xg,Yg,Vxg,Vyg,Magn=Plot_Field_Cylinder(X_S,Y_S,V_X,V_Y,False,2,0.6)
   # Show this particular step
   fig, ax = plt.subplots(figsize=(5, 3)) # This creates the figure
   # Or you can plot it as streamlines
   CL=plt.contourf(Xg,Yg,Magn,levels=np.arange(0,18,2))
   # One possibility is to use quiver
   STEPx=1;  STEPy=1
   plt.quiver(Xg[::STEPx,::STEPy],Yg[::STEPx,::STEPy],\
               Vxg[::STEPx,::STEPy],Vyg[::STEPx,::STEPy],color='k')
   plt.rc('text', usetex=True)      
   plt.rc('font', family='serif')
   plt.rc('xtick',labelsize=12)
   plt.rc('ytick',labelsize=12)
   plt.colorbar(CL)
   ax.set_aspect('equal') # Set equal aspect ratio
   ax.set_xlabel('$x[mm]$',fontsize=13)
   ax.set_ylabel('$y[mm]$',fontsize=13)
   ax.set_title('Tutorial 2: Cylinder Wake',fontsize=16)
   ax.set_xticks(np.arange(0,70,10))
   ax.set_yticks(np.arange(-10,11,10))
   ax.set_xlim([0,50])
   ax.set_ylim(-10,10)
   circle = plt.Circle((0,0),2.5,fill=True,color='r',edgecolor='k',alpha=0.5)
   plt.gcf().gca().add_artist(circle)
   plt.tight_layout()
   plt.show()
   Name[len(Name)-12:len(Name)]+' Plotted'
   return n_s, Xg, Yg, Vxg, Vyg, X_S, Y_S


def Plot_Field_Cylinder(X_S,Y_S,V_X,V_Y,PLOT,Step,Scale):
  # Number of n_X/n_Y from forward differences
   GRAD_X=np.diff(Y_S) 
   #GRAD_Y=np.diff(Y_S);
   # Depending on the reshaping performed, one of the two will start with
   # non-zero gradient. The other will have zero gradient only on the change.
   IND_X=np.where(GRAD_X!=0); DAT=IND_X[0]; n_y=DAT[0]+1;
   nxny=X_S.shape[0] # is the to be doubled at the end we will have n_s=2 * n_x * n_y
   #n_s=2*nxny
   # Reshaping the grid from the data
   n_x=(nxny//(n_y)) # Carefull with integer and float!
   Xg=np.transpose(X_S.reshape((n_x,n_y)))
   Yg=np.transpose(Y_S.reshape((n_x,n_y))) # This is now the mesh! 60x114.
   Mod=np.sqrt(V_X**2+V_Y**2)
   Vxg=np.transpose(V_X.reshape((n_x,n_y)))
   Vyg=np.transpose(V_Y.reshape((n_x,n_y)))
   Magn=np.transpose(Mod.reshape((n_x,n_y)))
  
   if PLOT:
   # Show this particular step
#    fig, ax = plt.subplots(figsize=(8, 5)) # This creates the figure
    plt.contourf(Xg,Yg,Magn)
    # One possibility is to use quiver
    STEPx=Step
    STEPy=Step
    plt.quiver(Xg[::STEPx,::STEPy],Yg[::STEPx,::STEPy],\
               Vxg[::STEPx,::STEPy],-Vyg[::STEPx,::STEPy],color='k',scale=Scale) # Create a quiver (arrows) plot
    
    plt.rc('text', usetex=True)      # This is Miguel's customization
    plt.rc('font', family='serif')
    plt.rc('xtick',labelsize=16)
    plt.rc('ytick',labelsize=16)
#    ax.set_aspect('equal') # Set equal aspect ratio
#    ax.set_xlabel('$x[mm]$',fontsize=16)
#    ax.set_ylabel('$y[mm]$',fontsize=16)
#    ax.set_title('Velocity Field via TR-PIV',fontsize=18)
#    ax.set_xticks(np.arange(0,40,10))
#    ax.set_yticks(np.arange(10,30,5))
#    ax.set_xlim([0,35])
#    ax.set_ylim(10,29)
#    ax.invert_yaxis() # Invert Axis for plotting purpose
#    plt.show()
       
   return Xg,Yg,Vxg,Vyg,Magn  
   


